from django.http import HttpResponseNotModified
from django.middleware import cache
from django.utils.cache import patch_cache_control, get_conditional_response, patch_response_headers
from django.utils.http import parse_http_date, parse_http_date_safe
from django.utils.timezone import now

from rmr.utils.patch import patch


class UpdateCacheMiddleware(cache.UpdateCacheMiddleware):
    """
    Do the same as the original one but try first to use 'key_prefix' from
    the request where it might be saved earlier

    see https://code.djangoproject.com/ticket/15855
    """

    def process_response(self, request, response):
        key_prefix = getattr(request, '_cache_key_prefix', self.key_prefix)
        cache_timeout = getattr(request, '_cache_cache_timeout', self.cache_timeout)

        if isinstance(response, HttpResponseNotModified):
            patch_response_headers(response, cache_timeout=cache_timeout)
            return response
        elif 'Expires' in response:
            # Replace 'max-age' value of 'Cache-Control' header by one
            # calculated from the 'Expires' header's date.
            # This is necessary because of Django's `FetchFromCacheMiddleware`
            # gets 'Cache-Control' header from the cache
            # where 'max-age' corresponds to the moment original response
            # was generated and thus may be already stale for the current time
            expires = parse_http_date(response['Expires'])
            timeout = expires - int(now().timestamp())
            patch_cache_control(response, max_age=timeout)

        with patch(self, 'key_prefix', key_prefix):
            with patch(self, 'cache_timeout', cache_timeout):
                return super().process_response(request, response)


class CacheMiddleware(cache.CacheMiddleware):
    """
    Despite of the original one this middleware supports callable 'key_prefix'
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if callable(self.key_prefix):
            self.key_function = self.key_prefix

    def key_function(self, request, *args, **kwargs):
        return self.key_prefix

    def get_key_prefix(self, request):
        key_prefix = self.key_function(
            request,
            *request.resolver_match.args,
            **request.resolver_match.kwargs
        )
        if key_prefix is not None:
            key_prefix = str(key_prefix).replace(' ', '_')
        return key_prefix

    def process_request(self, request):
        self.key_prefix = self.get_key_prefix(request)
        request._cache_cache_timeout = int(self.cache_timeout)
        response = super().process_request(request)
        if response:
            etag = response.get('ETag')
            last_modified = response.get('Last-Modified')
            if not (etag or last_modified):
                return response
            response = get_conditional_response(
                request,
                etag=etag,
                last_modified=last_modified and parse_http_date_safe(last_modified),
                response=response,
            )
            response['Last-Modified'] = last_modified
            return response

    def process_response(self, request, response):
        request._cache_key_prefix = self.get_key_prefix(request)

        # you must add rmr's UpdateCacheMiddleware at the top of the
        # MIDDLEWARE_CLASSES to be able to save responses in the cache
        # see https://code.djangoproject.com/ticket/15855
        return response



import logging


class Error(Exception):

    level = logging.NOTSET

    http_code = NotImplemented

    message = NotImplemented

    code = NotImplemented

    def __init__(self, message=None, *, code=None, level=None, http_code=None):
        super().__init__(message)
        self.message = message or self.message
        self.code = code or self.code
        self.level = level or self.level
        self.http_code = http_code or self.http_code

    def __str__(self):
        return '[{code}] {message}'.format(message=self.message, code=self.code)


class ServerError(Error):

    level = logging.ERROR

    http_code = 500


class ClientError(Error):

    level = logging.WARNING

    http_code = 400


def handle_error(error, exception=None):
    def _decorator(fn):
        def _fn(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except error:
                if exception:
                    raise exception
        return _fn
    return _decorator



"""
Django settings for Antares-Backend project.

Generated by 'django-admin startproject' using Django 1.8.2.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.8/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.8/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'secret_key'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = False

ALLOWED_HOSTS = []

CACHE_MIDDLEWARE_SECONDS = 0


# Application definition

INSTALLED_APPS = (
    'rmr',
)

MIDDLEWARE_CLASSES = (
    'rmr.extensions.middleware.cache.UpdateCacheMiddleware',
    'rmr.extensions.middleware.json.RequestDecoder',
)

ROOT_URLCONF = None


# Database
# https://docs.djangoproject.com/en/1.8/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'rmr_django',
    },
}


# Internationalization
# https://docs.djangoproject.com/en/1.8/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Logging
# https://docs.djangoproject.com/en/1.8/ref/settings/#id11

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'null': {
            'class': 'logging.NullHandler',
        },
    },
    'loggers': {
        'django': {
            'handlers': ['null'],
            'level': 'ERROR',
        },
        'rmr': {
            'handlers': ['null'],
            'level': 'ERROR',
        },
    },
}



import contextlib
from unittest import mock

import django.test
from django.db import models

from rmr.models.utils import BulkModelCreator


class SampleModel(models.Model):
    pass


class TestBulkModelCreator(django.test.TestCase):

    @mock.patch.object(SampleModel.objects, 'bulk_create')
    def test_no_bulk_create_on_exception(self, bulk_create):
        with contextlib.suppress(Exception):
            with BulkModelCreator() as creator:
                for i in range(0, 10):
                    creator.add(SampleModel())
                # Something goes wrong here:
                raise Exception
        self.assertFalse(bulk_create.called, 'bulk_create() was called, while should not!')



import json
import urllib.parse

from datetime import datetime
from unittest import mock

import django.test

from django import forms
from django.conf.urls import url
from django.core.urlresolvers import reverse
from django.test.utils import override_settings
from django.utils import timezone
from django.utils.decorators import method_decorator

from rmr.errors import ClientError, ServerError
from rmr.utils.test import data_provider, DataSet, Parametrized
from rmr.views import Json
from rmr.views.decorators.validation import validate_request


class JsonWithWarning(Json):

    def get(self, request):
        raise ClientError('WARNING_TEST_CASE', code='warning_test_case')


class JsonWithError(Json):

    def get(self, request):
        raise ServerError('ERROR_TEST_CASE', code='error_test_case')


class JsonWithoutError(Json):

    def get(self, request):
        pass


class CacheJson(Json):

    timestamp = None

    @classmethod
    def last_modified(cls, request, *args, **kwargs):
        if cls.timestamp:
            return datetime.fromtimestamp(cls.timestamp)
        return datetime(2015, 1, 1, tzinfo=timezone.utc)

    @classmethod
    def expires(cls):
        return 3600

    def get(self, request):
        pass

    def post(self, request):
        pass


class ValidationJson(Json):

    class GetValidationForm(forms.Form):
        get_required = forms.IntegerField()
        get_not_required = forms.IntegerField(required=False)

    class PostValidationForm(forms.Form):
        post_required = forms.IntegerField()
        post_not_required = forms.IntegerField(required=False)

    @method_decorator(validate_request(get=GetValidationForm))
    def get(self, request):
        pass

    @method_decorator(validate_request(post=PostValidationForm))
    def post(self, request):
        pass

    @method_decorator(validate_request(
        get=GetValidationForm,
        post=PostValidationForm,
    ))
    def patch(self, request):
        pass


urlpatterns = [
    url(r'warning', JsonWithWarning.as_view(), name='warning'),
    url(r'error', JsonWithError.as_view(), name='error'),
    url(r'ok', JsonWithoutError.as_view(), name='ok'),
    url(r'cache', CacheJson.as_view(), name='cache'),
    url(r'last_modified', CacheJson.as_view(), name='last_modified'),
    url(r'validate', ValidationJson.as_view(), name='validate'),
]


@override_settings(ROOT_URLCONF=__name__)
class JsonTestCase(django.test.TestCase, metaclass=Parametrized):

    @data_provider(
        DataSet(
            timestamp1=None,
            timestamp2=None,
            cache_stale_expected=False,
        ),
        DataSet(
            timestamp1=1420070400,
            timestamp2=1420070401,
            cache_stale_expected=True,
        ),
    )
    def test_last_modified_as_cache_key(self, timestamp1, timestamp2, cache_stale_expected):
        client = django.test.Client()

        mocked_get = mock.MagicMock(return_value=None)
        with mock.patch.multiple(CacheJson, timestamp=timestamp1, get=mocked_get):
            client.get(
                reverse('last_modified'),
            )
            self.assertTrue(mocked_get.called)

        mocked_get = mock.MagicMock(return_value=None)
        with mock.patch.multiple(CacheJson, timestamp=timestamp1, get=mocked_get):
            client.get(
                reverse('last_modified'),
            )
            self.assertFalse(mocked_get.called)

        mocked_get = mock.MagicMock(return_value=None)
        with mock.patch.multiple(CacheJson, timestamp=timestamp2, get=mocked_get):
            client.get(
                reverse('last_modified'),
            )
            self.assertEqual(cache_stale_expected, mocked_get.called)

        mocked_get = mock.MagicMock(return_value=None)
        with mock.patch.multiple(CacheJson, timestamp=timestamp2, get=mocked_get):
            client.get(
                reverse('last_modified'),
            )
            self.assertFalse(mocked_get.called)

    @data_provider(
        DataSet('warning', ClientError.http_code),
        DataSet('error', ServerError.http_code),
        DataSet('ok', Json.http_code),
    )
    def test_status_code(self, url_name, expected_status_code):
        client = django.test.Client()
        response = client.get(reverse(url_name))
        self.assertEqual(expected_status_code, response.status_code)

    def test_response_headers(self):
        client = django.test.Client()

        # GET request of not cached content with If-Modified-Since provided
        response = client.get(
            reverse('cache'),
            HTTP_IF_MODIFIED_SINCE='Thu, 01 Jan 2015 00:00:00 GMT',
        )
        self.assertIn('Content-Length', response)
        self.assertIn('Last-Modified', response)
        self.assertIn('Expires', response)
        self.assertIn('Cache-Control', response)
        self.assertIn('max-age=3600', response['Cache-Control'])
        self.assertEqual('Thu, 01 Jan 2015 00:00:00 GMT', response['Last-Modified'])
        self.assertEqual('0', response['Content-Length'])
        self.assertEqual(304, response.status_code)

        # GET request of cachable content
        response = client.get(reverse('cache'))
        self.assertEqual(Json.http_code, response.status_code)
        self.assertIn('Content-Length', response)
        self.assertIn('Content-Type', response)
        self.assertIn('Cache-Control', response)
        self.assertIn('Expires', response)
        self.assertIn('Last-Modified', response)
        self.assertEqual('14', response['Content-Length'])
        self.assertIn('max-age=3600', response['Cache-Control'])
        self.assertEqual('Thu, 01 Jan 2015 00:00:00 GMT', response['Last-Modified'])
        self.assertEqual('application/json', response['Content-Type'])

        # POST request of cachable content
        response = client.post(reverse('cache'))
        self.assertEqual(Json.http_code, response.status_code)
        self.assertIn('Content-Length', response)
        self.assertIn('Content-Type', response)
        self.assertNotIn('Expires', response)
        self.assertNotIn('Cache-Control', response)
        self.assertNotIn('Last-Modified', response)
        self.assertEqual('14', response['Content-Length'])
        self.assertEqual('application/json', response['Content-Type'])

        # GET request of cached content with If-Modified-Since provided
        response = client.get(
            reverse('cache'),
            HTTP_IF_MODIFIED_SINCE='Thu, 01 Jan 2015 00:00:00 GMT',
        )
        self.assertIn('Content-Length', response)
        self.assertIn('Last-Modified', response)
        self.assertIn('Expires', response)
        self.assertIn('Cache-Control', response)
        self.assertIn('max-age=3600', response['Cache-Control'])
        self.assertEqual('Thu, 01 Jan 2015 00:00:00 GMT', response['Last-Modified'])
        self.assertEqual('0', response['Content-Length'])
        self.assertEqual(304, response.status_code)

        # GET request of cached content without If-Modified-Since provided
        response = client.get(reverse('cache'))
        self.assertEqual(Json.http_code, response.status_code)
        self.assertIn('Content-Length', response)
        self.assertIn('Content-Type', response)
        self.assertIn('Cache-Control', response)
        self.assertIn('Expires', response)
        self.assertIn('Last-Modified', response)
        self.assertEqual('14', response['Content-Length'])
        self.assertIn('max-age=3600', response['Cache-Control'])
        self.assertEqual('Thu, 01 Jan 2015 00:00:00 GMT', response['Last-Modified'])
        self.assertEqual('application/json', response['Content-Type'])

        # POST request of cachable content with If-Modified-Since provided
        response = client.post(
            reverse('cache'),
            HTTP_IF_MODIFIED_SINCE='Thu, 01 Jan 2015 00:00:00 GMT',
        )
        self.assertEqual(Json.http_code, response.status_code)
        self.assertIn('Content-Length', response)
        self.assertNotIn('Expires', response)
        self.assertNotIn('Cache-Control', response)
        self.assertNotIn('Last-Modified', response)
        self.assertEqual('14', response['Content-Length'])
        self.assertEqual('application/json', response['Content-Type'])

        # GET request of non-cachable content
        response = client.get(reverse('ok'))
        self.assertIn('Cache-Control', response)
        self.assertIn('Content-Length', response)
        self.assertIn('Content-Type', response)
        self.assertIn('max-age=0', response['Cache-Control'])
        self.assertEqual('14', response['Content-Length'])
        self.assertEqual('application/json', response['Content-Type'])

    @data_provider(
        DataSet(
            offset=None,
            limit=None,
            limit_default=None,
            limit_max=None,
            expected=(0, None),
        ),
        DataSet(
            offset=5,
            limit=10,
            limit_default=None,
            limit_max=None,
            expected=(5, 15),
        ),
        DataSet(
            offset='5',
            limit='10',
            limit_default=None,
            limit_max=None,
            expected=(5, 15),
        ),
        DataSet(
            offset=None,
            limit=10,
            limit_default=None,
            limit_max=None,
            expected=(0, 10),
        ),
        DataSet(
            offset=5,
            limit=None,
            limit_default=None,
            limit_max=None,
            expected=(5, None),
        ),
        DataSet(
            offset=None,
            limit=None,
            limit_default=10,
            limit_max=None,
            expected=(0, 10),
        ),
        DataSet(
            offset=5,
            limit=None,
            limit_default=10,
            limit_max=None,
            expected=(5, 15),
        ),
        DataSet(
            offset=5,
            limit=None,
            limit_default=10,
            limit_max=None,
            expected=(5, 15),
        ),
        DataSet(
            offset=5,
            limit=10,
            limit_default=None,
            limit_max=10,
            expected=(5, 15),
        ),
    )
    def test_get_range(self, offset, limit, limit_default, limit_max, expected):
        self.assertEqual(
            expected,
            Json.get_range(
                offset=offset,
                limit=limit,
                limit_default=limit_default,
                limit_max=limit_max,
            ),
        )

    @data_provider(
        DataSet(
            offset=None,
            limit=-1,
            limit_default=None,
            limit_max=None,
            expected_error_code='incorrect_limit',
        ),
        DataSet(
            offset=-1,
            limit=None,
            limit_default=None,
            limit_max=None,
            expected_error_code='incorrect_offset',
        ),
        DataSet(
            offset='blah',
            limit=None,
            limit_default=None,
            limit_max=None,
            expected_error_code='incorrect_limit_or_offset',
        ),
        DataSet(
            offset=None,
            limit='blah',
            limit_default=None,
            limit_max=None,
            expected_error_code='incorrect_limit_or_offset',
        ),
        DataSet(
            offset=None,
            limit=11,
            limit_default=None,
            limit_max=10,
            expected_error_code='max_limit_exceeded',
        ),
        DataSet(
            offset=None,
            limit=None,
            limit_default=11,
            limit_max=10,
            expected_error_code='max_limit_exceeded',
        ),
        DataSet(
            offset=None,
            limit=None,
            limit_default=None,
            limit_max=10,
            expected_error_code='limit_not_provided',
        ),
    )
    def test_get_range_errors(self, offset, limit, limit_default, limit_max, expected_error_code):
        with self.assertRaises(ClientError) as error:
            Json.get_range(
                offset=offset,
                limit=limit,
                limit_default=limit_default,
                limit_max=limit_max,
            )
        self.assertEqual(expected_error_code, error.exception.code)

    def test_validate_request(self):
        client = django.test.Client()
        path = reverse('validate')

        # GET validation
        response = client.get(path, data=dict(
            get_required=123,
            get_not_required=123,
            unknown=123,
        ))
        self.assertEqual(
            200,
            response.status_code,
        )

        # POST validation
        response = client.post(
            path,
            data=json.dumps(dict(
                post_required=123,
                post_not_required=123,
                unknown=123,
            )),
            content_type='application/json',
        )
        self.assertEqual(
            200,
            response.status_code,
        )

        # GET and POST validation
        response = client.patch(
            '{path}?{query}'.format(
                path=path,
                query=urllib.parse.urlencode(dict(get_required=123)),
            ),
            data=json.dumps(dict(post_required=123)),
            content_type='application/json',
        )
        self.assertEqual(
            200,
            response.status_code,
        )

    @data_provider(
        DataSet(
            method='GET',
            query={},
            data={},
            invalid_params={'get_required'},
        ),
        DataSet(
            method='POST',
            query={},
            data={},
            invalid_params={'post_required'},
        ),
        DataSet(
            method='PATCH',
            query={},
            data={},
            invalid_params={'post_required', 'get_required'},
        ),
        DataSet(
            method='PATCH',
            query=dict(get_not_required='wrong_value'),
            data=dict(post_not_required='wrong_value'),
            invalid_params={'post_required', 'get_required', 'get_not_required', 'post_not_required'},
        ),
    )
    def test_validate_request_errors(self, method, query, data, invalid_params):
        client = django.test.Client()
        path = reverse('validate')
        response = client.generic(
            method=method,
            path='{path}?{query}'.format(
                path=path,
                query=urllib.parse.urlencode(query),
            ),
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(
            400,
            response.status_code,
        )
        data = json.loads(response.content.decode())
        actual_invalid_params = data.get('error', {}).get('description', {}).keys()
        self.assertSetEqual(invalid_params, set(actual_invalid_params))



"""
Range fields placed here are actual only for Django <= 1.8

Django 1.9 already has 'upper' and 'lower' lookups for the range fields
"""

from django.contrib.postgres import fields
from django.db import models


class DateRangeField(fields.DateRangeField):
    pass


class DateTimeRangeField(fields.DateTimeRangeField):
    pass


class FloatRangeField(fields.FloatRangeField):
    pass


class BigIntegerRangeField(fields.BigIntegerRangeField):
    pass


class IntegerRangeField(fields.IntegerRangeField):
    pass


class Lower(models.Transform):
    lookup_name = 'lower'

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return 'lower({lhs})'.format(lhs=lhs), params


class Upper(models.Transform):
    lookup_name = 'upper'

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return 'upper({lhs})'.format(lhs=lhs), params


@DateRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.DateField()


@DateRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.DateField()


@DateTimeRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.DateTimeField()


@DateTimeRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.DateTimeField()


@FloatRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.FloatField()


@FloatRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.FloatField()


@BigIntegerRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.BigIntegerField()


@BigIntegerRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.BigIntegerField()


@IntegerRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.IntegerField()


@IntegerRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.IntegerField()



import django.http
import django.test

from django.conf.urls import url
from django.test.utils import override_settings

import rmr.views

from rmr.utils.test import data_provider, DataSet, Parametrized

urlpatterns = [
    url(r'', lambda request: django.http.JsonResponse(request.POST)),
]


@override_settings(ROOT_URLCONF=__name__)
class RequestDecoderTestCase(django.test.TestCase, metaclass=Parametrized):

    def setUp(self):
        self.client = django.test.Client()

    def test_wrong_charset(self):
        response = self.client.post(
            '/',
            data=b'\xd0',
            content_type='application/json',
        )
        self.assertIsInstance(response, django.http.HttpResponseBadRequest)
        self.assertEqual(b'bad unicode', response.content)

    @data_provider(
        DataSet(data='"'),
        DataSet(data='""'),
        DataSet(data='[]'),
        DataSet(data='null'),
        DataSet(data='{'),
        DataSet(data='hello'),
        DataSet(data='hello=world'),
    )
    def test_wrong_json(self, data):
        response = self.client.post(
            '/',
            data=data,
            content_type='application/json',
        )
        self.assertIsInstance(response, django.http.HttpResponseBadRequest)
        self.assertEqual(b'malformed data', response.content)

    @data_provider(
        DataSet(data='{"a": 1}'),
        DataSet(data='{}'),
    )
    def test_valid_json(self, data):
        response = self.client.post(
            '/',
            data=data,
            content_type='application/json',
        )
        self.assertIsInstance(response, django.http.JsonResponse)
        self.assertEqual(rmr.views.Json.http_code, response.status_code)
        self.assertEqual(response.content.decode('utf-8'), data)



import unittest

from rmr.utils.iterate import split_every


class TestUtilsTestCase(unittest.TestCase):

    def test_split_every(self):
        expected_result = [[0, 1], [2, 3], [4]]
        result = [list(x) for x in split_every(range(5), 2)]
        self.assertEqual(expected_result, result)



from collections import defaultdict


class BulkModelCreator:
    """
    BulkModelCreator accumulates instances of any Django models and saves them into the db (using bulk_create)
    when count of particular model instances reaches 'batch_size' parameter.
    On exit from context manager will create all instances, that left in collection.

    Example:

    with BulkModelCreator(100) as creator:
        for row in csv_reader.read():
            instance = get_instance(row)
            creator.add(instance)

    """

    def __init__(self, batch_size=300):
        self.batches = defaultdict(list)
        self.batch_size = batch_size

    def add(self, instance):
        self.batches[instance.__class__].append(instance)
        if len(self.batches[instance.__class__]) == self.batch_size:
            self.flush(instance.__class__)
            return True

    def flush(self, model_class):
        model_class.objects.bulk_create(
            self.batches[model_class],
            self.batch_size
        )
        self.batches[model_class] = []

    def flush_all(self):
        for key in self.batches:
            self.flush(key)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        exc_type is None and self.flush_all()



from datetime import datetime

from django.utils import timezone


def fromtimestamp(timestamp):
    return datetime.fromtimestamp(timestamp, tz=timezone.get_current_timezone())


def strptime(date_string, format, default_tz=timezone.utc):
    dt = datetime.strptime(date_string, format)
    if timezone.is_naive(dt):
        dt = timezone.make_aware(dt, timezone=default_tz)
    return timezone.localtime(dt)



from django import forms


class MultiValueField(forms.TypedMultipleChoiceField):

    def valid_value(self, value):
        return True


class BooleanField(forms.Field):

    empty_values = {None}

    def to_python(self, value):
        if value in (True, False):
            return value
        return None


class OffsetLimit(forms.Form):

    offset = forms.IntegerField(required=False, min_value=0)
    limit = forms.IntegerField(required=False)

    @classmethod
    def with_limit_required(cls, *, limit_max_value):
        limit = forms.IntegerField(
            required=True,
            max_value=limit_max_value,
        )
        return type('OffsetLimitRequired', (cls, ), dict(limit=limit))
